<?php namespace SokoWeb\PersistentData;

use Generator;
use SokoWeb\Database\Query\Modify;
use SokoWeb\Database\Query\Select;
use SokoWeb\Interfaces\Database\IConnection;
use SokoWeb\Interfaces\Database\IAuditLogger;
use SokoWeb\Interfaces\Database\IResultSet;
use SokoWeb\Interfaces\PersistentData\IPersistentDataManager;
use SokoWeb\PersistentData\Model\Model;
use SokoWeb\PersistentData\Model\ModelWithSlug;

class PersistentDataManager implements IPersistentDataManager
{
    const SLUG_MAX_LENGTH = 255;

    private IConnection $dbConnection;

    private ?IAuditLogger $auditLogger;

    public function __construct(IConnection $dbConnection, ?IAuditLogger $auditLogger = null)
    {
        $this->dbConnection = $dbConnection;
        $this->auditLogger = $auditLogger;
    }

    public function selectFromDb(Select $select, string $type, bool $useRelations = false, array $withRelations = [])
    {
        $select = $this->createSelect($select, $type, $useRelations, $withRelations);

        $data = $select->execute()->fetch(IResultSet::FETCH_ASSOC);

        if ($data === null) {
            return null;
        }

        $model = new $type();
        $this->fillWithData($data, $model, $withRelations);

        return $model;
    }

    public function selectMultipleFromDb(Select $select, string $type, bool $useRelations = false, array $withRelations = []): Generator
    {
        $select = $this->createSelect($select, $type, $useRelations, $withRelations);
        $result = $select->execute();

        while ($data = $result->fetch(IResultSet::FETCH_ASSOC)) {
            $model = new $type();
            $this->fillWithData($data, $model, $withRelations);

            yield $model;
        }
    }

    public function selectFromDbById($id, string $type, bool $useRelations = false, array $withRelations = [])
    {
        $select = new Select($this->dbConnection);
        $select->whereId($id);

        return $this->selectFromDb($select, $type, $useRelations, $withRelations);
    }

    public function selectFromDbBySlug(string $slug, string $type, bool $useRelations = false, array $withRelations = [])
    {
        $select = new Select($this->dbConnection);
        $select->where('slug', '=', $slug);

        return $this->selectFromDb($select, $type, $useRelations, $withRelations);
    }

    public function loadRelationsFromDb(Model $model, bool $recursive = false, array $withRelations = []): void
    {
        $relations = $model::getRelations();
        if (count($withRelations)) {
            $relations = array_intersect_key($relations, array_flip($withRelations));
        }

        foreach ($relations as $relation => $relationType) {
            $camel = str_replace('_', '', ucwords($relation, '_'));

            $methodGet = 'get' . $camel . 'Id';
            $methodSet = 'set' . $camel;

            $relationId = $model->$methodGet();

            if ($relationId !== null) {
                $relationModel = $this->selectFromDbById($relationId, $relationType, $recursive, $withRelations);

                $model->$methodSet($relationModel);
            }
        }
    }

    public function saveToDb(Model $model): void
    {
        $this->syncRelations($model);

        $modified = $model->toArray();
        $id = $model->getId();

        $modify = new Modify($this->dbConnection, $model::getTable(), $this->auditLogger);

        if ($id !== null) {
            $original = $model->getSnapshot();
            $diff = [];

            foreach ($original as $key => $value) {
                if ($value === $modified[$key]) {
                    unset($modified[$key]);
                } else {
                    $diff[$key] = ['old' => $value, 'new' => $modified[$key]];
                }
            }

            if (count($modified) > 0) {
                if ($model instanceof ModelWithSlug && isset($modified['slug'])) {
                    $diff['slug']['new'] = $modified['slug'] = $this->generateUniqueSlug($model, $modified['slug']);
                }

                $modify->setId($id);
                $modify->setDiff($diff);
                $modify->fill($modified);
                $modify->save();
            }
        } else {
            if ($model instanceof ModelWithSlug) {
                $slug = $model->generateSlug();
                $modified['slug'] = $this->generateUniqueSlug($model, $slug);
            }

            $modify->fill($modified);
            $modify->save();

            $model->setId($modify->getId());
        }

        $model->saveSnapshot();
    }

    public function deleteFromDb(Model $model): void
    {
        $modify = new Modify($this->dbConnection, $model::getTable(), $this->auditLogger);
        $modify->setId($model->getId());
        $modify->fill($model->toArray());
        $modify->delete();

        $model->setId(null);
        $model->resetSnapshot();
    }

    private function generateUniqueSlug(ModelWithSlug $model, string $notUniqueSlug): string
    {
        $numbered = 1;

        do {
            if ($numbered > 1) {
                $slug = substr($notUniqueSlug, 0, static::SLUG_MAX_LENGTH - (strlen((string)$numbered) + 1));
                $slug = $notUniqueSlug . '_' . (string)$numbered;
            } else {
                $slug = substr($notUniqueSlug, 0, static::SLUG_MAX_LENGTH);
            }

            $select = new Select($this->dbConnection, $model::getTable());
            $select->where('slug', '=', $slug);

            $numbered++;
        } while ($select->count() != 0);

        $model->setSlug($slug);

        return $slug;
    }

    private function createSelect(Select $select, string $type, bool $useRelations = false, array $withRelations = []): Select
    {
        $table = call_user_func([$type, 'getTable']);
        $fields = call_user_func([$type, 'getFields']);

        $columns = [];

        foreach ($fields as $field) {
            $columns[] = [$table, $field];
        }

        $select->from($table);

        if ($useRelations) {
            $relations = call_user_func([$type, 'getRelations']);
            if (count($withRelations)) {
                $relations = array_intersect_key($relations, array_flip($withRelations));
            }

            $columns = array_merge($columns, $this->getRelationColumns($table, $relations, $withRelations));

            $this->leftJoinRelations($select, $table, $table, $relations, $withRelations);
            $select->columns($columns);
        } else {
            $select->columns($columns);
        }

        return $select;
    }

    private function getRelationColumns(string $table, array $relations, array $withRelations): array
    {
        $columns = [];

        foreach ($relations as $relation => $relationType) {
            $relationTableAlias = $table . '__' . $relation;
            $relationTable = call_user_func([$relationType, 'getTable']);
            foreach (call_user_func([$relationType, 'getFields']) as $relationField) {
                $columns[] = [$relationTableAlias, $relationField, $relation . '__' . $relationField];
            }

            $relationsOfRelation = call_user_func([$relationType, 'getRelations']);
            if (count($withRelations)) {
                $relationsOfRelation = array_intersect_key($relationsOfRelation, array_flip($withRelations));
            }
            $columns = array_merge($columns, $this->getRelationColumns($relationTable, $relationsOfRelation, $withRelations));
        }

        return $columns;
    }

    private function leftJoinRelations(Select $select, string $table, string $tableAlias, array $relations, array $withRelations): void
    {
        foreach ($relations as $relation => $relationType) {
            $relationTableAlias = $table . '__' . $relation;
            $relationTable = call_user_func([$relationType, 'getTable']);
            $select->setTableAliases([$relationTableAlias => $relationTable]);
            $select->leftJoin($relationTableAlias, [$relationTableAlias, 'id'], '=', [$tableAlias, $relation . '_id']);

            $relationsOfRelation = call_user_func([$relationType, 'getRelations']);
            if (count($withRelations)) {
                $relationsOfRelation = array_intersect_key($relationsOfRelation, array_flip($withRelations));
            }
            $this->leftJoinRelations($select, $relationTable, $relationTableAlias, $relationsOfRelation, $withRelations);
        }
    }

    private function fillWithData(array &$data, Model $model, array $withRelations = [], ?string $modelKey = null): void
    {
        $relations = $model::getRelations();
        if (count($withRelations)) {
            $relations = array_intersect_key($relations, array_flip($withRelations));
        }

        while (key($data)) {
            $key = key($data);
            $value = current($data);
            $relation = key($relations);

            if (strpos($key, '__') === false) {
                $method = 'set' . str_replace('_', '', ucwords($key, '_'));

                if (method_exists($model, $method) && isset($value)) {
                    $model->$method($value);
                }

                next($data);
            } else if (isset($modelKey) && substr($key, 0, strlen($modelKey . '__')) === $modelKey . '__') {
                $key = substr($key, strlen($modelKey) + 2);

                $method = 'set' . str_replace('_', '', ucwords($key, '_'));

                if (method_exists($model, $method) && isset($value)) {
                    $model->$method($value);
                }

                next($data);
            } else if (substr($key, 0, strlen($relation . '__')) === $relation . '__') {
                if ($data[$relation . '__id'] !== null) {
                    $relationType = current($relations);
                    $relationModel = new $relationType();
                    $this->fillWithData($data, $relationModel, $withRelations, $relation);

                    $method = 'set' . str_replace('_', '', ucwords($relation, '_'));
                    $model->$method($relationModel);
                } else {
                    while (substr($key, 0, strlen($relation . '__')) === $relation . '__') {
                        next($data);
                        $key = key($data);
                    }
                }

                next($relations);
            } else {
                return;
            }
        }

        $model->saveSnapshot();
    }

    private function syncRelations(Model $model): void
    {
        foreach ($model::getRelations() as $relation => $relationType) {
            $camel = str_replace('_', '', ucwords($relation, '_'));

            $methodGet = 'get' . $camel;
            $methodSet = 'set' . $camel . 'Id';

            $relationModel = $model->$methodGet();

            if ($relationModel !== null) {
                $model->$methodSet($relationModel->getId());
            }
        }
    }
}
